@@ -1,5 +1,7 @@
 //! Example how to use pure `egui_glow` without [`epi`].
 
+#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")] // hide console window on Windows in release
+
 fn create_display(
     event_loop: &glutin::event_loop::EventLoop<()>,
 ) -> (
@@ -42,13 +44,13 @@ fn main() {
     let event_loop = glutin::event_loop::EventLoop::with_user_event();
     let (gl_window, gl) = create_display(&event_loop);
 
-    let mut egui_glow = egui_glow::EguiGlow::new(&gl_window, &gl);
+    let mut egui_glow = egui_glow::EguiGlow::new(gl_window.window(), &gl);
 
     event_loop.run(move |event, _, control_flow| {
         let mut redraw = || {
             let mut quit = false;
 
-            let (needs_repaint, shapes) = egui_glow.run(gl_window.window(), |egui_ctx| {
+            let needs_repaint = egui_glow.run(gl_window.window(), |egui_ctx| {
                 egui::SidePanel::left("my_side_panel").show(egui_ctx, |ui| {
                     ui.heading("Hello World!");
                     if ui.button("Quit").clicked() {
@@ -76,7 +78,7 @@ fn main() {
 
                 // draw things behind egui here
 
-                egui_glow.paint(&gl_window, &gl, shapes);
+                egui_glow.paint(gl_window.window(), &gl);
 
                 // draw things on top of egui here
 
